﻿using System.Buffers;
using System.Runtime.Intrinsics;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Json;

namespace histcopy
{
    public class Histcopy(string source, string destination, CancellationToken token)
    {

        #region 変数
        // 必須設定
        private readonly string SourceDirectoryName = @"\\?\" + source;
        private readonly string RootDirectoryName = @"\\?\" + destination;
        private readonly string RecentDirectoryName = @"\\?\" + Path.Combine(destination, "recent");
        private readonly string HistoryDirectoryName = @"\\?\" + Path.Combine(destination, $"{DateTime.Now:yyyy-MM-dd}");
        
        // 任意設定
        public int HeaderLength { get; init; } = 4096;
        public int FooterLength { get; init; } = 4096;
        public int BufferLength { get; init; } = 1024 * 1024;
        public double VerifyRate { get; init; } = 0.01;

        // 制御
        private readonly CancellationToken Token = token;
        private readonly CountdownEvent CountDown = new(1);
        private readonly Dictionary<FileKey, List<FileInfo>> Moves = [];
        private readonly List<FileInfo?> Copies = [];
        private StreamWriter? LogWriter;

        // 統計
        private int CopyCount;
        private long CopyLength;
        private int DirectoryMoveCount;
        private int FileMoveCount;
        private int UnmatchCount;
        private int NameRecycleCount;
        private int DataRecycleCount;
        private int DirectoryCreateCount;
        private int DirectoryDeleteCount;
        private int SkipCount;
        private int DirectoryReadCount;
        private int FileReadCount;
        private long ReadLength;
        private int FailCount;
        private long FileLength;
        #endregion

        #region 進捗
        public event EventHandler<ProgressEventArgs>? Progress;

        public class ProgressEventArgs(string message) : EventArgs
        {
            public string Message { get; } = message;
        }
        #endregion

        #region 実行
        public void Run()
        {
            // ソース存在チェック
            var sourceDirectory = new DirectoryInfo(SourceDirectoryName);
            if (!sourceDirectory.Exists)
                throw new DirectoryNotFoundException($"{SourceDirectoryName} は見つかりません。");

            // ログ作成
            var historyDirectory = new DirectoryInfo(HistoryDirectoryName);
            if (!historyDirectory.Exists)
            {
                historyDirectory.Create();
                DirectoryCreateCount++;
            }
            using var logWriter = new StreamWriter(Path.Combine(HistoryDirectoryName, "histcopy_log.txt"), true);
            LogWriter = logWriter;
            var beginTime = DateTime.Now;
            WriteLog($"begin {beginTime:yyyy-MM-dd HH:mm:ss}");

            // ディレクトリスキャン、移動
            Task.Run(() => Scan(sourceDirectory));
            CountDown.Wait();

            // ファイル移動（リサイクル）
            Parallel.For(0, Copies.Count, copyIndex =>
            {
                try
                {
                    var sourceFile = Copies[copyIndex];
                    if (sourceFile == null || Token.IsCancellationRequested)
                        return;
                    var relativeFileName = GetRelativeName(sourceFile, SourceDirectoryName);
                    var recentFileName = Path.Combine(RecentDirectoryName, relativeFileName);
                    if (Moves.TryGetValue(new FileKey(sourceFile.LastWriteTime, sourceFile.Length), out var list))
                    {
                        lock (list)
                        {
                            var foundIndex = -1;
                            var isData = false;
                            for (var index = 0; index < list.Count; index++)
                            {
                                if (list[index].Name == sourceFile.Name)
                                {
                                    foundIndex = index;
                                    break;
                                }
                            }
                            if (foundIndex < 0)
                            {
                                for (var index = 0; index < list.Count; index++)
                                {
                                    if (CompareHeaderFooter(list[index], sourceFile))
                                    {
                                        foundIndex = index;
                                        isData = true;
                                        break;
                                    }
                                }
                            }
                            if (foundIndex >= 0)
                            {
                                var movedFile = list[foundIndex];
                                movedFile.MoveTo(recentFileName);
                                if (isData)
                                    Interlocked.Increment(ref DataRecycleCount);
                                else
                                    Interlocked.Increment(ref NameRecycleCount);
                                list.RemoveAt(foundIndex);
                                Copies[copyIndex] = null;
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    WriteLog(ex.Message);
                    Interlocked.Increment(ref FailCount);
                }
            });
            Moves.Clear();
            Moves.TrimExcess();

            // 履歴にある空のディレクトリを削除
            CountDown.Reset();
            Task.Run(() => DeleteEmptyDirectory(historyDirectory, null));
            CountDown.Wait();
            CountDown.Dispose();

            // ファイルコピー
            Copies.Sort((file1, file2) => (file2 == null ? "" : file2.FullName).CompareTo(file1 == null ? "" : file1.FullName));
            for (var copyIndex = Copies.Count - 1; copyIndex >= 0 && !Token.IsCancellationRequested; copyIndex--)
            {
                try
                {
                    var sourceFile = Copies[copyIndex];
                    Copies.RemoveAt(copyIndex);
                    if (sourceFile != null)
                    {
                        var relativeFileName = GetRelativeName(sourceFile, SourceDirectoryName);
                        var recentFileName = Path.Combine(RecentDirectoryName, relativeFileName);
                        WriteLog(relativeFileName);
                        sourceFile.CopyTo(recentFileName);
                        CopyCount++;
                        CopyLength += sourceFile.Length;
                    }
                }
                catch (Exception ex)
                {
                    WriteLog(ex.Message);
                    FailCount++;
                }
            }
            Copies.Clear();
            Copies.TrimExcess();

            // ベリファイ
            using var verifyState = new VerifyState(this, Path.Combine(RootDirectoryName, "verify.json"), SourceDirectoryName);
            var leftLength = (long)Math.Ceiling(FileLength * VerifyRate);
            if (leftLength > 0 && !Token.IsCancellationRequested)
                WriteLog("verify");
            while (verifyState.Read() && leftLength > 0 && !Token.IsCancellationRequested)
            {
                try
                {
                    // ファイル取得
                    var sourceFile = verifyState.File ?? throw new NullReferenceException();
                    var relativeFileName = GetRelativeName(sourceFile, SourceDirectoryName);
                    var recentFile = new FileInfo(Path.Combine(RecentDirectoryName, relativeFileName));
                    WriteLog(relativeFileName);

                    bool isDataMatch;
                    {
                        // ファイルオープン
                        using var sourceStream = sourceFile.OpenRead();
                        FileReadCount++;
                        using var recentStream = recentFile.OpenRead();
                        FileReadCount++;
                        leftLength -= sourceFile.Length;
                        isDataMatch = sourceFile.Length == recentFile.Length;

                        // バッファセット初期化
                        using var readBuffer = new VerifyBufferSet();
                        using var compareBuffer = new VerifyBufferSet();

                        while (!Token.IsCancellationRequested && isDataMatch)
                        {
                            // 読み込みバッファ取得
                            readBuffer.Source = ArrayPool<byte>.Shared.Rent(BufferLength);
                            readBuffer.Recent = ArrayPool<byte>.Shared.Rent(BufferLength);

                            // 非同期で読み込み開始
                            using var sourceTask = sourceStream.ReadAsync(readBuffer.Source, 0, BufferLength, CancellationToken.None);
                            using var recentTask = recentStream.ReadAsync(readBuffer.Recent, 0, BufferLength, CancellationToken.None);

                            // 前回読み込んだデータを比較
                            isDataMatch = compareBuffer.Compare();
                            compareBuffer.Return();

                            // 読み込み完了を待機
                            Task.WaitAll(sourceTask, recentTask);

                            // 読み込んだ長さを統計値に加算
                            readBuffer.SourceLength = sourceTask.Result;
                            ReadLength += readBuffer.SourceLength;
                            readBuffer.RecentLength = recentTask.Result;
                            ReadLength += readBuffer.RecentLength;

                            // すべてのデータを読み込んだときは抜ける
                            if (readBuffer.SourceLength == 0 && readBuffer.RecentLength == 0)
                                break;

                            // 読み込んだデータを比較用バッファセットに移動
                            readBuffer.MoveTo(compareBuffer);
                        }
                    }

                    // データ不一致のとき、再コピー
                    if (!isDataMatch)
                    {
                        // 履歴ディレクトリを作成、ユニークな名前を生成・追加
                        var relativeDirectoryName = GetRelativeName(sourceFile.Directory ?? throw new NullReferenceException(), SourceDirectoryName);
                        var historyDirectory2 = new HistoryDirectory(this, Path.Combine(HistoryDirectoryName, relativeDirectoryName));
                        var uniqueName = historyDirectory2.CreateAndAdd(recentFile.Name);

                        // 最近ファイルを移動
                        File.Move(recentFile.FullName, Path.Combine(historyDirectory2.FullName, uniqueName));
                        UnmatchCount++;

                        // コピー
                        sourceFile.CopyTo(recentFile.FullName);
                        CopyCount++;
                        CopyLength += sourceFile.Length;
                    }
                }
                catch (Exception ex)
                {
                    WriteLog(ex.Message);
                    FailCount++;
                }
            }

            // 結果表示
            var endTime = DateTime.Now;
            WriteLog($"end {endTime:yyyy-MM-dd HH:mm:ss}");
            var span = endTime - beginTime;
            WriteLog("time " + (span.TotalMinutes >= 1.0
                ? $"{span.Hours + span.Days * 24:00}:{span.Minutes:00}:{span.Seconds:00}"
                : $"{span.TotalSeconds:N3} second"));
            WriteLog("copy", CopyCount, CopyLength);
            WriteLog("move", DirectoryMoveCount, "directory");
            WriteLog("move", FileMoveCount, "file");
            WriteLog("move", UnmatchCount, "file (unmatch)");
            WriteLog("recycle", NameRecycleCount, "file (name)");
            WriteLog("recycle", DataRecycleCount, "file (data)");
            WriteLog("create", DirectoryCreateCount, "directory");
            WriteLog("delete", DirectoryDeleteCount, "directory");
            WriteLog("skip", SkipCount, "file");
            WriteLog("read", DirectoryReadCount, "directory");
            WriteLog("read", FileReadCount, ReadLength);
            WriteLog("fail", FailCount, "item");
            WriteLog("");
        }
        #endregion

        #region ソースディレクトリをスキャン
        private void Scan(DirectoryInfo sourceDirectory)
        {
            try
            {
                // キャンセルされたときは抜ける
                if (Token.IsCancellationRequested)
                    return;

                // ディレクトリを取得
                var relativeDirectoryName = GetRelativeName(sourceDirectory, SourceDirectoryName);
                var recentDirectory = new DirectoryInfo(Path.Combine(RecentDirectoryName, relativeDirectoryName));
                var historyDirectory = new HistoryDirectory(this, Path.Combine(HistoryDirectoryName, relativeDirectoryName));

                // 最近ディレクトリが存在しない場合は作成
                var recentDirectoryExists = recentDirectory.Exists;
                if (!recentDirectoryExists)
                {
                    recentDirectory.Create();
                    Interlocked.Increment(ref DirectoryCreateCount);
                }

                // ディレクトリ情報を取得
                var sourceInfos = sourceDirectory.GetFileSystemInfos().ToDictionary(info => info.Name.ToUpper());
                Interlocked.Increment(ref DirectoryReadCount);
                var recentInfos = recentDirectoryExists ? recentDirectory.GetFileSystemInfos().ToDictionary(info => info.Name.ToUpper()) : [];
                Interlocked.Increment(ref DirectoryReadCount);

                // ソースを処理
                foreach (var sourceInfo in sourceInfos.Values)
                {
                    if (sourceInfo is DirectoryInfo sourceSubDirectory)
                    {
                        // ソースがサブディレクトリとき、スキャン
                        CountDown.AddCount(1);
                        Task.Run(() => Scan(sourceSubDirectory));
                    }
                    else
                    {
                        // ソースがファイルのとき
                        var sourceFile = (FileInfo)sourceInfo;
                        Interlocked.Add(ref FileLength, sourceFile.Length);
                        if (!recentInfos.TryGetValue(sourceFile.Name.ToUpper(), out FileSystemInfo? recentInfo))
                        {
                            // 最近にファイル・ディレクトリが存在しないとき、ファイルコピー
                            lock (Copies)
                                Copies.Add(sourceFile);
                        }
                        else
                        {
                            if (recentInfo is DirectoryInfo recentSubDirectory)
                            {
                                // 最近にディレクトリが存在するとき、ディレクトリ移動、ファイルコピー
                                AddMoveAndCopy(recentSubDirectory, historyDirectory, sourceFile);
                            }
                            else
                            {
                                // 最近にファイルが存在し、ソースファイルの属性と一致しないとき、ファイル移動、ファイルコピー
                                var recentFile = (FileInfo)recentInfo;
                                if (sourceFile.Name != recentFile.Name ||
                                    sourceFile.Length != recentFile.Length ||
                                    sourceFile.LastWriteTime != recentFile.LastWriteTime)
                                {
                                    AddMoveAndCopy(recentFile, historyDirectory, sourceFile);
                                }
                                else
                                {
                                    Interlocked.Increment(ref SkipCount);
                                }
                            }
                        }
                    }
                }

                // 最近にファイル・ディレクトリが存在し、ソースにファイル・ディレクトリが存在しないとき、移動
                foreach (var recentInfo in recentInfos.Values)
                    if (!sourceInfos.ContainsKey(recentInfo.Name.ToUpper()))
                        AddMoveAndCopy(recentInfo, historyDirectory, null);
            }
            catch (Exception ex)
            {
                WriteLog(ex.Message);
                Interlocked.Increment(ref FailCount);
            }
            finally
            {
                CountDown.Signal();
            }
        }
        #endregion

        #region ファイル・ディレクトリを移動、ファイルコピー
        private void AddMoveAndCopy(FileSystemInfo recentInfo, HistoryDirectory historyDirectory, FileInfo? sourceFile)
        {
            CountDown.AddCount(1);
            Task.Run(() => {
                try
                {
                    // キャンセルされたときは抜ける
                    if (Token.IsCancellationRequested)
                        return;

                    // 履歴ディレクトリを作成、ユニークな名前を生成・追加
                    var uniqueName = historyDirectory.CreateAndAdd(recentInfo.Name);
                    var destination = Path.Combine(historyDirectory.FullName, uniqueName);
                    var relativeName = GetRelativeName(recentInfo, RecentDirectoryName);

                    if (recentInfo is DirectoryInfo recentSubDirectory)
                    {
                        // 最近ディレクトリを移動
                        recentSubDirectory.MoveTo(destination);
                        Interlocked.Increment(ref DirectoryMoveCount);

                        // ディレクトリを再帰スキャンし、ファイルを移動リストに追加
                        CountDown.AddCount(1);
                        Task.Run(() => AddMoves(recentSubDirectory));
                    }
                    else if (recentInfo is FileInfo recentFile)
                    {
                        // 最近ファイルを移動
                        recentFile.MoveTo(destination);
                        Interlocked.Increment(ref FileMoveCount);

                        // ファイルを移動リストに追加
                        AddMoves(recentFile);
                    }

                    // コピーリストに追加
                    if (sourceFile != null)
                        lock (Copies)
                            Copies.Add(sourceFile);
                }
                catch (Exception ex)
                {
                    WriteLog(ex.Message);
                    Interlocked.Increment(ref FailCount);
                }
                finally
                {
                    CountDown.Signal();
                }
            });
        }
        #endregion

        #region 空ディレクトリを削除
        private void DeleteEmptyDirectory(DirectoryInfo directory, DirectoryItem? parent)
        {
            try
            {
                // キャンセルされたときは抜ける
                if (Token.IsCancellationRequested)
                    return;

                // ディレクトリ情報を読み取り、項目を生成
                var infos = directory.GetFileSystemInfos();
                Interlocked.Increment(ref DirectoryReadCount);
                var item = new DirectoryItem(directory, parent, infos.Length);

                if (item.Count == 0)
                {
                    // ディレクトリが空のときは上位ディレクトリも含めて削除
                    while (item.Count == 0)
                    {
                        item.Directory.Delete();
                        Interlocked.Increment(ref DirectoryDeleteCount);
                        if (item.Parent == null)
                            break;
                        item = item.Parent;
                        Interlocked.Decrement(ref item.Count);
                    }
                }
                else
                {
                    foreach (var subDirectory in infos.OfType<DirectoryInfo>())
                    {
                        // サブディレクトリをスキャン
                        CountDown.AddCount(1);
                        Task.Run(() => DeleteEmptyDirectory(subDirectory, item));
                    }
                }
            }
            catch (Exception ex)
            {
                WriteLog(ex.Message);
                Interlocked.Increment(ref FailCount);
            }
            finally
            {
                CountDown.Signal();
            }
        }

        private class DirectoryItem(DirectoryInfo directory, DirectoryItem? parent, int count)
        {
            public readonly DirectoryInfo Directory = directory;
            public readonly DirectoryItem? Parent = parent;
            public int Count = count;
        }
        #endregion

        #region 移動ファイルを追加
        private void AddMoves(FileInfo file)
        {
            var key = new FileKey(file.LastWriteTime, file.Length);
            lock (Moves)
            {
                if (!Moves.TryGetValue(key, out var list))
                {
                    list = [];
                    Moves.Add(key, list);
                }
                list.Add(file);
            }
        }

        private void AddMoves(DirectoryInfo directory)
        {
            try
            {
                // キャンセルされたときは抜ける
                if (Token.IsCancellationRequested)
                    return;

                var infos = directory.GetFileSystemInfos();
                Interlocked.Increment(ref DirectoryReadCount);

                foreach (var info in infos)
                {
                    if (info is DirectoryInfo subDirectory)
                    {
                        // サブディレクトリをスキャン
                        CountDown.AddCount(1);
                        Task.Run(() => AddMoves(subDirectory));
                    }
                    else if (info is FileInfo file)
                    {
                        // 移動ファイルを追加
                        AddMoves(file);
                    }
                }
            }
            catch (Exception ex)
            {
                WriteLog(ex.Message);
                Interlocked.Increment(ref FailCount);
            }
            finally
            {
                CountDown.Signal();
            }
        }

        private record FileKey(DateTime Date, long Length);
        #endregion

        #region 履歴ディレクトリ
        private class HistoryDirectory(Histcopy histCopy, string fullName)
        {
            private readonly Histcopy HistCopy = histCopy;
            public readonly string FullName = fullName;
            private HashSet<string>? Keys;

            public string CreateAndAdd(string baseName)
            {
                lock (this)
                {
                    if (Keys == null)
                    {
                        var directory = new DirectoryInfo(FullName);
                        lock (HistCopy) // dir1 と dir1\dir2 の作成を全体で排他
                        {
                            if (!directory.Exists)
                            {
                                directory.Create();
                                Interlocked.Increment(ref HistCopy.DirectoryCreateCount);
                                Keys = [];
                            }
                        }
                        if (Keys == null)
                        {
                            var infos = directory.GetFileSystemInfos();
                            Interlocked.Increment(ref HistCopy.DirectoryReadCount);
                            Keys = (from fsi in infos select fsi.Name.ToUpper()).ToHashSet();
                        }
                    }
                    var name = baseName;
                    var index = 0;
                    while (!Keys.Add(name.ToUpper()))
                    {
                        name = $"{baseName}.{index++}";
                    }
                    return name;
                }
            }
        }
        #endregion

        #region ヘッダ・フッタを比較
        private bool CompareHeaderFooter(FileInfo file1, FileInfo file2)
        {
            // サイズチェック
            if (file1.Length != file2.Length)
                return false;

            if (file1.Length > 0 && (HeaderLength > 0 || FooterLength > 0))
            {
                // ファイルオープン
                using var stream1 = file1.OpenRead();
                FileReadCount++;
                using var stream2 = file2.OpenRead();
                FileReadCount++;

                // ヘッダ比較
                if (!CompareFilePart(stream1, stream2, 0, HeaderLength))
                    return false;

                // フッタ比較
                if (file1.Length > HeaderLength && !CompareFilePart(stream1, stream2, Math.Max(HeaderLength, file1.Length - FooterLength), FooterLength))
                    return false;
            }

            return true;
        }

        private bool CompareFilePart(FileStream stream1, FileStream stream2, long offset, int length)
        {
            if (length == 0)
                return true;

            byte[]? buffer1 = null;
            byte[]? buffer2 = null;
            try
            {
                stream1.Seek(offset, SeekOrigin.Begin);
                buffer1 = ArrayPool<byte>.Shared.Rent(length);
                var readLength1 = stream1.Read(buffer1, 0, length);
                ReadLength += readLength1;

                stream2.Seek(offset, SeekOrigin.Begin);
                buffer2 = ArrayPool<byte>.Shared.Rent(length);
                var readLength2 = stream2.Read(buffer2, 0, length);
                ReadLength += readLength2;

                return CompareBuffer(buffer1, buffer2, readLength1, readLength2);
            }
            finally
            {
                if (buffer1 != null)
                    ArrayPool<byte>.Shared.Return(buffer1);
                if (buffer2 != null)
                    ArrayPool<byte>.Shared.Return(buffer2);
            }
        }
        #endregion

        #region バッファ内容を比較
        private static bool CompareBuffer(byte[] buffer1, byte[] buffer2, int length1, int length2)
        {
            if (length1 != length2)
                return false;
            if (Vector512.IsHardwareAccelerated && length1 >= 64)
            {
                for (int index = 0, length64 = (length1 - 1) & 0x7fff_ffc0; index < length64; index += 64)
                    if (Vector512.Create(buffer1, index) != Vector512.Create(buffer2, index))
                        return false;
                return Vector512.Create(buffer1, length1 - 64) == Vector512.Create(buffer2, length1 - 64);
            }
            if (Vector256.IsHardwareAccelerated && length1 >= 32)
            {
                for (int index = 0, length32 = (length1 - 1) & 0x7fff_ffe0; index < length32; index += 32)
                    if (Vector256.Create(buffer1, index) != Vector256.Create(buffer2, index))
                        return false;
                return Vector256.Create(buffer1, length1 - 32) == Vector256.Create(buffer2, length1 - 32);
            }
            for (var index = 0; index < length1; index++)
                if (buffer1[index] != buffer2[index])
                    return false;
            return true;
        }
        #endregion

        #region 相対名を取得
        private static string GetRelativeName(FileSystemInfo info, string directoryName)
        {
            var name = info.FullName;
            var index = directoryName.Length;
            if (!directoryName.EndsWith('\\'))
                index++;
            return index >= name.Length ? "" : name[index..];
        }
        #endregion

        #region ログに書き込み
        private void WriteLog(string message)
        {
            if (LogWriter != null)
                lock (LogWriter)
                    try { LogWriter.WriteLine(message); } catch { }
            if (message.Length > 0)
                try { Progress?.Invoke(this, new ProgressEventArgs(message)); } catch { }
        }

        private void WriteLog(string message, int count, string itemName)
        {
            if (count != 0)
                WriteLog($"{message} {count:N0} {itemName}");
        }

        private void WriteLog(string message, int count, long length)
        {
            if (count != 0)
                WriteLog($"{message} {count:N0} file {length / 1024.0:N0} KB");
        }
        #endregion
                
        #region ベリファイ状態
        private class VerifyState : IDisposable
        {
            private readonly Histcopy Histcopy;
            private readonly string JsonFileName;
            private readonly string DirectoryName;
            private DirectoryItem Directory;
            private bool IsBeginFirst;
            public FileInfo? File;

            public VerifyState(Histcopy histcopy, string jsonFileName, string directoryName)
            {
                Histcopy = histcopy;
                JsonFileName = jsonFileName;
                DirectoryName = directoryName;
                Directory = new DirectoryItem(new DirectoryInfo(DirectoryName), null);

                try
                {
                    // JSONファイルが存在しないとき、抜ける
                    var jsonFile = new FileInfo(jsonFileName);
                    if (!jsonFile.Exists)
                    {
                        IsBeginFirst = true;
                        return;
                    }

                    // JSONファイルを読み込み
                    using var stream = jsonFile.OpenRead();
                    var serializer = new DataContractJsonSerializer(typeof(JsonData));
                    var data = (JsonData?)serializer.ReadObject(stream);
                    if (data == null || data.FileName == null)
                        throw new NullReferenceException();

                    // パスを分解し、名前キューを作成
                    var nameQueue = new Queue<string>(data.FileName.Split('\\'));
                    var name = nameQueue.Dequeue();
                    while (Directory.Queue.Count > 0)
                    {
                        var info = Directory.Queue.Peek();
                        var result = info.Name.CompareTo(name);

                        // ディレクトリのとき、次の階層へ移動する
                        if (result == 0 && nameQueue.Count > 0 && info is DirectoryInfo subDirectory)
                        {
                            Directory.Queue.Dequeue();
                            Directory = new DirectoryItem(subDirectory, Directory);
                            name = nameQueue.Dequeue();
                            continue;
                        }

                        // ファイルのとき、名前が過ぎたとき（見つからないとき）は抜ける
                        if (result >= 0)
                            break;

                        // 次の名前へ移動
                        Directory.Queue.Dequeue();
                    }
                }
                catch (Exception ex)
                {
                    Histcopy.WriteLog(ex.Message);
                    Histcopy.FailCount++;
                }
            }

            public bool Read()
            {
                while (true)
                {
                    if (Directory.Queue.Count > 0)
                    {
                        var info = Directory.Queue.Dequeue();
                        if (info is FileInfo file)
                        {
                            // キュー項目がファイルのとき、カレント項目をセット、true（データあり）を返す
                            File = file;
                            return true;
                        }
                        try
                        {
                            // キュー項目がディレクトリのとき、次の階層へ移動する。読み取れないときは次の項目へ移動する。
                            Directory = new DirectoryItem((DirectoryInfo)info, Directory);
                        }
                        catch (Exception ex)
                        {
                            Histcopy.WriteLog(ex.Message);
                            Histcopy.FailCount++;
                        }
                        continue;
                    }
                    if (Directory.Parent != null)
                    {
                        // 親ディレクトリがあるとき、上の階層へ移動
                        Directory = Directory.Parent;
                        continue;
                    }
                    // 親ディレクトリがないとき（すべての項目をスキャンし終わったとき）
                    if (IsBeginFirst)
                    {
                        // 最初からスキャン済みのとき、false（データなし）を返す
                        File = null;
                        return false;
                    }
                    // 最初からスキャンし直し
                    Directory = new DirectoryItem(new DirectoryInfo(DirectoryName), null);
                    IsBeginFirst = true;
                }
            }

            public void Dispose()
            {
                try
                {
                    var jsonFile = new FileInfo(JsonFileName);
                    if (jsonFile.Exists)
                        jsonFile.Delete();
                    if (File != null)
                    {
                        using var stream = jsonFile.Create();
                        var serializer = new DataContractJsonSerializer(typeof(JsonData));
                        serializer.WriteObject(stream, new JsonData() { FileName = GetRelativeName(File, DirectoryName) }); ;
                    }
                }
                catch (Exception ex)
                {
                    Histcopy.WriteLog(ex.Message);
                    Histcopy.FailCount++;
                }
            }

            [DataContract]
            private class JsonData
            {
                [DataMember] public string? FileName;
            }

            private class DirectoryItem(DirectoryInfo directory, DirectoryItem? parent)
            {
                public readonly DirectoryInfo Directory = directory;
                public readonly DirectoryItem? Parent = parent;
                public readonly Queue<FileSystemInfo> Queue = new(from info in directory.GetFileSystemInfos() orderby info.Name select info);
            }
        }
        #endregion

        #region ベリファイバッファセット
        private class VerifyBufferSet : IDisposable
        {
            public byte[]? Source;
            public byte[]? Recent;
            public int SourceLength;
            public int RecentLength;

            public void MoveTo(VerifyBufferSet destination)
            {
                destination.Source = Source;
                destination.Recent = Recent;
                destination.SourceLength = SourceLength;
                destination.RecentLength = RecentLength;
                Source = null;
                Recent = null;
                SourceLength = 0;
                RecentLength = 0;
            }

            public void Return()
            {
                if (Source != null)
                {
                    ArrayPool<byte>.Shared.Return(Source);
                    Source = null;
                }
                if (Recent != null)
                {
                    ArrayPool<byte>.Shared.Return(Recent);
                    Recent = null;
                }
                SourceLength = 0;
                RecentLength = 0;
            }

            public void Dispose()
            {
                Return();
            }

            public bool Compare()
            {
                if (Source == null || Recent == null)
                    return true;
                return CompareBuffer(Source, Recent, SourceLength, RecentLength);
            }

        }
        #endregion

    }
}
